//This file is a part of xboa
//
//xboa is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//xboa is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with xboa in the doc folder.  If not, see 
//<http://www.gnu.org/licenses/>.
//


/*
Hitcore.h
Designed to improve speed. My concern here is to provide C get
functions so that other code (i.e. Bunchcore) can do the lookup
directly rather than going through Python

Contains hitcore object, which holds most of the data for the hit
get functions
  get internal data, either directly or using a string, int or python
  string to reference the internal data
  two sets of functions, one to return double, one to return int
  if only I had c++ templates!
set functions
  set internal data, either directly or using a string, int or python
  string to reference the internal data
  two sets of functions, one to return double, one to return int
  if only I had c++ templates!
global_weight_table
  this is a dictionary that maps (spill, event_number, particle_number) to
  (global_weight)

Also I define a vector object which is a variable length array of void*.
This is designed to look more-or-less like a std::vector from C++, but much less safe i.e. has:
  data_length - memory initialised), 
  allocated_length - memory reserved but not necessarily initialised)
  size_t - size of the objects in the vector
  data* - pointer to the data
Initialisation and allocation via vector_init, vector_alloc
Data insertion via vector_insert
Element referencing via vector_el and vector_el_bc (with bound checking)


*/

#ifndef Hitcore_h 
#define Hitcore_h


#include <Python.h>

//Hitcore data struct
typedef struct {
    PyObject_HEAD;
    double x;
    double y;
    double z;
    double t;
    double px;
    double py;
    double pz;
    double energy;
    double weight;
    int event; 
    double mass;
    double bx;
    double by;
    double bz;
    double ex;
    double ey;
    double ez;
    double sx;
    double sy;
    double sz; 
    double pl;
    double tau;
    double e_dep;
    double charge;
    int station;
    int pid;
    int status;
    int particle;
    int spill;
} Hitcore;
typedef double (*hc_get_dbl_function)(Hitcore* self);
typedef void   (*hc_set_dbl_function)(Hitcore* self, double value);

typedef int    (*hc_get_int_function)(Hitcore* self);
typedef void   (*hc_set_int_function)(Hitcore* self, int value);

//vector
typedef struct {void* data; size_t data_length; size_t allocd_length; size_t object_size; } vector;


#ifdef Hitcore_c
//deallocate memory from Hitcore
static void hc_Hitcore_dealloc(Hitcore * self);
static void hc_Hitcore_free(Hitcore * self);
//initialise Hitcore objects
static PyObject *hc_Hitcore_alloc(PyTypeObject *self, Py_ssize_t nitems);
static PyObject *hc_Hitcore_new(PyTypeObject *self, Py_ssize_t nitems);
static int  hc_Hitcore_init(PyObject* self, PyObject *args, PyObject *kwds);


////// GET METHODS ///////
//specialist get_methods
double hc_get_x(Hitcore* self);
double hc_get_y(Hitcore* self);
double hc_get_z(Hitcore* self);
double hc_get_t(Hitcore* self);
double hc_get_px(Hitcore* self);
double hc_get_py(Hitcore* self);
double hc_get_pz(Hitcore* self);
double hc_get_energy(Hitcore* self);
double hc_get_mass(Hitcore* self);
double hc_get_bx(Hitcore* self);
double hc_get_by(Hitcore* self);
double hc_get_bz(Hitcore* self);
double hc_get_ex(Hitcore* self);
double hc_get_ey(Hitcore* self);
double hc_get_ez(Hitcore* self);
double hc_get_sx(Hitcore* self);
double hc_get_sy(Hitcore* self);
double hc_get_sz(Hitcore* self);
double hc_get_pl(Hitcore* self); //pl is path length
double hc_get_tau(Hitcore* self); //tau is proper time
double hc_get_e_dep(Hitcore* self); //e_dep is energy deposited
double hc_get_charge(Hitcore* self);

//Get weight
double hc_get_total_weight (Hitcore* self);
double hc_get_local_weight (Hitcore* self);
double hc_get_global_weight(Hitcore* self);
// expose dictionary (really only for testing)
PyObject* hc_get_global_weight_dict(PyObject* self, PyObject* args);
// dictionary key
PyObject* weight_dict_key(Hitcore* key);

//Integer gets
int    hc_get_event_number   (Hitcore* self);
int    hc_get_station        (Hitcore* self);
int    hc_get_pid            (Hitcore* self);
int    hc_get_status         (Hitcore* self);
int    hc_get_particle_number(Hitcore* self);
int    hc_get_spill          (Hitcore* self);
int    hc_get_unique_id      (Hitcore* self);

//mapping of strings to functions
//e.g. hc_get_dbl_function_by_int   (1)  (some_hitcore)        - returns some_hitcore.y
//e.g. hc_get_dbl_function_by_string("y")(some_hitcore)        - returns some_hitcore.y

static hc_get_dbl_function  hc_get_dbl_function_by_string(const char* variable);

static hc_get_int_function  hc_get_int_function_by_string(const char* variable);

//general get methods
//for ints
//static int       hc_get_int_by_int   (Hitcore* self, int index, int* ierr);
static int       hc_get_int_by_string(Hitcore* self, const char* variable, int* ierr);
//for doubles
//static double    hc_get_dbl_by_int   (Hitcore* self, int index, int* ierr);
static double    hc_get_dbl_by_string(Hitcore* self, const char* variable, int* ierr);
static PyObject* hc_get_python(PyObject* self, PyObject* args); //ONLY returns DOUBLES
///////////////////////////

////// SET METHODS ////////
//specialist set_methods
void hc_set_x(Hitcore* self, double value);
void hc_set_y(Hitcore* self, double value);
void hc_set_z(Hitcore* self, double value);
void hc_set_t(Hitcore* self, double value);
void hc_set_px(Hitcore* self, double value);
void hc_set_py(Hitcore* self, double value);
void hc_set_pz(Hitcore* self, double value);
void hc_set_energy(Hitcore* self, double value);
void hc_set_mass(Hitcore* self, double value);
void hc_set_bx(Hitcore* self, double value);
void hc_set_by(Hitcore* self, double value);
void hc_set_bz(Hitcore* self, double value);
void hc_set_ex(Hitcore* self, double value);
void hc_set_ey(Hitcore* self, double value);
void hc_set_ez(Hitcore* self, double value);
void hc_set_sx(Hitcore* self, double value);
void hc_set_sy(Hitcore* self, double value);
void hc_set_sz(Hitcore* self, double value);
void hc_set_pl(Hitcore* self, double value); //pl is path length
void hc_set_tau(Hitcore* self, double value); //tau is proper time
void hc_set_e_dep(Hitcore* self, double value); //e_dep is energy deposited
void hc_set_charge(Hitcore* self, double value);

//Integer sets
void hc_set_event_number   (Hitcore* self, int value);
void hc_set_station        (Hitcore* self, int value);
void hc_set_pid            (Hitcore* self, int value);
void hc_set_status         (Hitcore* self, int value);
void hc_set_particle_number(Hitcore* self, int value);
void hc_set_spill          (Hitcore* self, int value);

// update id
void update_unique_id(Hitcore* self);

//Set weight
void hc_set_local_weight (Hitcore* self, double weight);
void hc_set_global_weight(Hitcore* self, double weight);

//mapping of strings to functions
//e.g. hc_set_dbl_function_by_int   (1)  (some_hitcore, value) - sets y to value
//e.g. hc_set_dbl_function_by_string("y")(some_hitcore, value) - sets y to value
//doubles
static hc_set_dbl_function  hc_set_dbl_function_by_string(const char* variable);
//static void                 hc_set_dbl_by_int(Hitcore* self, int index, double value, int* ierr);
static void                 hc_set_dbl_by_string(Hitcore* self, const char* variable, double value, int* ierr);
//ints
static hc_set_int_function  hc_set_int_function_by_string(const char* variable);
static void                 hc_set_int_by_string         (Hitcore* self, const char* variable, double value, int* ierr);

//general set functions
static PyObject* hc_set_variables_to_python(PyObject* self, PyObject* args);
static PyObject* hc_set_python(PyObject* self, PyObject* args);
static PyObject* hc_clear_global_weights(PyObject* self, PyObject* args);

//static check that data integrity is ok (at compile time)
static PyObject* hc_integrity_test(PyObject* self, PyObject* args);
//initialise Hitcore python module
PyMODINIT_FUNC   hc_initHitcore(void);

//vector is like an std::vector - but without iterators (assume you'll just use ints)
//note no bound checking! be careful...
//data is a pointer to the array; 
//data_length is the length of the data in the array (i.e. sizeof(vector))
//allocd length is the allocated length of the array
//object_size is the size of objects in the array
//typedef above
//typedef struct {void* data; size_t data_length; size_t allocd_length; size_t object_size; } vector;
//allocate a new vector - object_size must be > 0
static vector* vector_alloc   (size_t object_size, int n_elements);
//realloc the vec
static vector* vector_realloc (vector* vec, int n_elements);
//build a vector from allocated data - data must be allocated elsewhere
static vector* vector_init    (void* data, size_t data_length, size_t allocd_length, size_t object_size);
//delete the vector
static void    vector_free (vector* vec);
//insert elements found between target_start and target_end into vec at the point vec->data+insert_point
//if insert_start == NULL, inserts at end (appends)
//fail if runs out of memory
static void  vector_insert    (vector* vec, void* start, void* end, size_t insert_point, int* ierr);
//return vector element at point int
static void* vector_el        (vector* vec, int element);
//return vector element at point int - with bound checking, returns NULL on out-of-bounds
static void* vector_el_bc     (vector* vec, int element);
//return number of elements
static size_t   vector_size      (vector* vec) 
{return vec->data_length/vec->object_size;}

//// NOT IMPLEMENTED YET:
/*
//remove elements found between erase_start and erase_end, keeping vec contiguous
static void  vector_remove    (vector* vec, void* start, void* end);
//sort between sort_start and sort_end
//if sort_start == NULL use vec_start; if sort_end == NULL use vec_end
static void  vector_sort      (vector* vec, void* start, void* end);
//find pointer to the highest value of vec elements that has vec[i] <= find_key
//where comparator returns <0 for a < b; 0 for a==b; >0 for a>b
static void* vector_find_lower(vector* vec, void* start, void* end, void* find_key, int (*comparator)(void* a, void* b));
*/

// does nothing (compiler warning issue)
void no_warn(void);

//global weights
PyObject* global_weight_table;
int last_id;
#endif //Hitcore_c

#endif //Hitcore_h

